import csv
import os
import copy
import math
import random
import pickle

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
from datetime import datetime
from datetime import timedelta
from datetime import time
# from sklearn.neural_network import MLPClassifier
# from sklearn.ensemble import RandomForestClassifier

#import tensorflow as tf
import keras
from keras.models import Sequential
from keras.layers import Dense, Dropout, Activation, Flatten
from keras.optimizers import SGD
# from imblearn.keras import BalancedBatchGenerator
# from imblearn.under_sampling import NearMiss
from sklearn.preprocessing import MinMaxScaler

from keras.layers import Conv2D
from keras.layers import MaxPooling2D

from keras.datasets import cifar10

from numpy import moveaxis

from keras.models import Model
from keras.layers import Input, LSTM

from functools import wraps
# from keras.utils.vis_utils import plot_model

from scipy.signal import filtfilt, butter


from matplotlib import rcParams
from keras.regularizers import l2

# rcParams['font.family'] = 'Times New Roman'
# rcParams['font.size'] = 32
# plt.rc('xtick', labelsize=32) 
# plt.rc('ytick', labelsize=32)
# plt.rcParams['text.usetex'] = True
# plt.rcParams["font.weight"] = "bold"
# plt.rc('font', weight='bold')

import argparse
import sys



def memoize(read_cache = True, cache_name = 'default'):
    def _memorize(function):
        @wraps(function)
        def __memorize(*args, **kwargs):
            if read_cache == True:
                print('read '+cache_name+ ' from cache!')
                cache = pickle.load(open(cache_name, 'rb'))
                return cache
            else:
                cur = function(*args, **kwargs)
                pickle.dump(cur, open(cache_name, "wb"))
                return cur
        return __memorize
    return _memorize

class Weather():

    def __init__(self):
        '''
        station: specify a weather station from station list
        test_feature_list: specify which weather parameter want to be predicted

        '''
        self.version = "7.0"
        self.version_info = "error check, time zone mismatch"
        self.data_path_0 = "alabama/"
        self.neighbor_dat_path_0 = "alabama_neighbor/"
        self.wrf_data_path_0 = "wrf/"
        self.data_path = None
        self.wrf_data_path = None
        # self.station_list = ['andalusia', 'atmore','bayminette','castleberry','dixie','elberta','fairhope','florala', 'foley', 'gasque','geneva','jay','kinston','leakesville','loxley','mobiledr','mobileusaw','saraland']
        self.station_list = ['elberta','fairhope','florala', 'foley', 'gasque','geneva','jay','kinston','leakesville','loxley','mobiledr','mobileusaw','saraland']

        # self.exp = ['Precip_TB3_Tot','Temp_C','WndDir_2m','WndDir_10m','WndSpd_2m','WndSpd_10m','WndSpd_Vert','WndSpd_2m_Max','WndSpd_10m_Max','WndSpd_Vert_Max','WndSpd_Vert_Min','WndSpd_2m_Std','WndSpd_10m_Std','WndSpd_2m_WVc_1','WndSpd_2m_WVc_2', 'WndSpd_2m_WVc_3','WndSpd_2m_WVc_4','WndSpd_10m_WVc_1','WndSpd_10m_WVc_2','WndSpd_10m_WVc_3','WndSpd_10m_WVc_4','WndSpd_Vert_Tot']
        
        self.exp = ['Precip_TB3_Tot','WndDir_2m','WndDir_10m','WndSpd_2m','WndSpd_10m','WndSpd_Vert']
        
        #self.feature_list_full = ['Year','DayOfMon','DayOfYear','Hour','Minute','PTemp', "IRTS_Trgt", 'Precip_TX_Tot', 'IRTS_Body', "SoilSfcT", "SoilT_5cm","SoilT_10cm","SoilT_20cm","SoilT_50cm","SoilT_100cm","AirT_1pt5m","AirT_2m","AirT_9pt5m","AirT_10m","RH_2m","RH_10m","Pressure_1","Pressure_2","TotalRadn","QuantRadn","WndDir_2m","WndDir_10m","WndSpd_2m","WndSpd_10m","WndSpd_Vert","Vitel_100cm_a","Vitel_100cm_b","Vitel_100cm_c","Vitel_100cm_d","WndSpd_2m_Max","WndSpd_10m_Max","WndSpd_Vert_Max","WndSpd_Vert_Min","WndSpd_2m_Std","WndSpd_10m_Std","SoilType","eR","el","Temp_C","er_tc","wfv", "NaCl","SoilCond","SoilCond_tc","SoilWaCond_tc", "WndSpd_2m_WVc_1", "WndSpd_2m_WVc_2","WndSpd_2m_WVc_3","WndSpd_2m_WVc_4","WndSpd_10m_WVc_1", "WndSpd_10m_WVc_2","WndSpd_10m_WVc_3","WndSpd_10m_WVc_4","WndSpd_Vert_Tot"]
        #self.feature_list_full_value = ['PTemp', "IRTS_Trgt", 'Precip_TX_Tot', 'IRTS_Body', "SoilSfcT", "SoilT_5cm","SoilT_10cm","SoilT_20cm","SoilT_50cm","SoilT_100cm","AirT_1pt5m","AirT_2m","AirT_9pt5m","AirT_10m","RH_2m","RH_10m","Pressure_1","Pressure_2","TotalRadn","QuantRadn","WndDir_2m","WndDir_10m","WndSpd_2m","WndSpd_10m","WndSpd_Vert","Vitel_100cm_a","Vitel_100cm_b","Vitel_100cm_c","Vitel_100cm_d","WndSpd_2m_Max","WndSpd_10m_Max","WndSpd_Vert_Max","WndSpd_Vert_Min","WndSpd_2m_Std","WndSpd_10m_Std","SoilType","eR","el","Temp_C","er_tc","wfv", "NaCl","SoilCond","SoilCond_tc","SoilWaCond_tc", "WndSpd_2m_WVc_1", "WndSpd_2m_WVc_2","WndSpd_2m_WVc_3","WndSpd_2m_WVc_4","WndSpd_10m_WVc_1", "WndSpd_10m_WVc_2","WndSpd_10m_WVc_3","WndSpd_10m_WVc_4","WndSpd_Vert_Tot"]
        
        self.feature_list_full = ['Year','DayOfMon','DayOfYear','Hour','Minute','Temp_C','PTemp','IRTS_Body', "Vitel_100cm_a","Vitel_100cm_b","Vitel_100cm_d","SoilCond", "SoilWaCond_tc", "eR", "wfv","RH_10m"]
        self.feature_list_full_value = ['Temp_C','PTemp',  'IRTS_Body', "Vitel_100cm_a","Vitel_100cm_b","Vitel_100cm_d","SoilCond", "SoilWaCond_tc", "eR", "wfv","RH_10m"]
        
        # self.wrf_feature_list_full = ['Year','Month','Date','Hour','Minute','temperature', 'windspd_ground',  'windspd_10m', 'u_500hpa', 'v_500hpa', 'u_850hpa', 'v_850hpa', 'u_1000hpa', 'v_1000hpa', 'u_10m', 'v_10m','humi','cloud_base_pressure','cloud_top_pressure','3000_h','1000_h','ground_mis']
        # self.feature_list_full = ["AirT_2m", "Precip_TX_Tot", "WndSpd_10m", "WndDir_10m","RH_2m","RH_10m",'WndDir_2m', 'WndSpd_2m']
        # self.wrf_feature_list_full = ['Year','Month','Date','Hour','Minute','u_500hpa', 'v_500hpa', 'cloud_base_pressure','cloud_top_pressure','3000_h']
        self.wrf_feature_list_full = ['Year','Month','Date','Hour','Minute','u_500hpa', 'v_500hpa', 'cloud_base_pressure','cloud_top_pressure','3000_h','1000_h','ground_mis', 'u_10m','v_10m','u_250hpa','v_250hpa','u_80m','v_80m','surface_pressure']
        # self.wrf_feature_list_value = ['temperature', 'windspd_ground',  'windspd_10m', 'u_500hpa', 'v_500hpa', 'u_850hpa', 'v_850hpa', 'u_1000hpa', 'v_1000hpa', 'u_10m', 'v_10m','humi','cloud_base_pressure','cloud_top_pressure','3000_h','1000_h','ground_mis']
        # self.wrf_feature_list_value = ['u_500hpa', 'v_500hpa', 'cloud_base_pressure','cloud_top_pressure','3000_h']
        self.wrf_feature_list_value = ['u_500hpa', 'v_500hpa', 'cloud_base_pressure','cloud_top_pressure','3000_h','1000_h','ground_mis', 'u_10m','v_10m','u_250hpa','v_250hpa','u_80m','v_80m','surface_pressure']
        self.test_feature_list = ["PTemp"]
        self.lat_lon = [None, None]

        self.wrf = True
        self.mode = ""

        # self.exp_save_path = self.station+'_NN_no.model'

        # data
        self.encoder_input_data = None
        self.decoder_input_data = None
        self.decoder_target_data = None

        # Lstm model
        self.encoder_feature_size = len(self.feature_list_full_value)
        self.decoder_feature_size = len(self.test_feature_list)+2   # add 2 control bit
        self.latent_dim = 256
        self.batch_size = 640  # Batch size for training.
        self.epochs = 50  # Number of epochs to train for.

        #  Define an input sequence and process it.
        self.encoder_inputs = Input(shape=(None, self.encoder_feature_size))
        self.encoder_lstm = LSTM(self.latent_dim, return_state=True)
        self.encoder_outputs, self.state_h, self.state_c = self.encoder_lstm(self.encoder_inputs)
        # We discard `encoder_outputs` and only keep the states.
        self.encoder_states = [self.state_h, self.state_c]

        # Set up the decoder, using `encoder_states` as initial state.
        self.decoder_inputs = Input(shape=(None, self.decoder_feature_size))
        # We set up our decoder to return full output sequences,
        # and to return internal states as well. We don't use the
        # return states in the training model, but we will use them in inference.
        self.decoder_lstm = LSTM(self.latent_dim, return_sequences=True, return_state=True)
        self.decoder_outputs, _, _ = self.decoder_lstm(self.decoder_inputs, initial_state = self.encoder_states)
        self.decoder_dense = Dense(self.decoder_feature_size, activation='sigmoid')
        self.decoder_outputs = self.decoder_dense(self.decoder_outputs)

        
        # Define the model that will turn
        # `encoder_input_data` & `decoder_input_data` into `decoder_target_data`
        self.model = Model([self.encoder_inputs, self.decoder_inputs], self.decoder_outputs)
       
        self.encoder_model = None
        self.decoder_model = None

  
    def read_lstm_file(self, year = '2017', year_during = 1, start_date = "0101", during = 1, shift = None, seq = 5):
        print("Reading Mesonet File")
        last_year = False
        train_y_all = pd.DataFrame(columns = self.test_feature_list)

        name_list = self.feature_list_full
        train_x_all = pd.DataFrame(columns=self.feature_list_full_value)

        # alabama data
        for yd in range(year_during):    # year
            if yd == year_during-1:
                last_year = True
            else:
                last_year = False

            curr_year = str(int(year)+yd)
            self.data_path = self.data_path_0 + curr_year+'/'
            year_folders = os.listdir(self.data_path)


            train_y = pd.DataFrame(columns = self.test_feature_list)
            train_x = pd.DataFrame(columns = name_list)
            
            file_count = 0
            read_flag = False

            for f_alabma in year_folders:
                
                
                if start_date and not read_flag:
                    if not f_alabma[4:] == str(start_date):
                        continue
                    read_flag=True


                record_path = self.data_path+f_alabma+'/'
                station_records = os.listdir(record_path)

                for r in station_records:
                    
                    r_split = r.split('_')
                    if not self.station == r_split[1]:
                        continue

                    station_record = record_path+r
                    try:
                        station_data = pd.read_csv(station_record)
                    except:
                        continue

                    # print(station_data)


                    station_data_t = copy.deepcopy(station_data)

                    for column in station_data_t.columns:
                        if not column in self.test_feature_list:
                            del station_data_t[column]


                    for column in station_data.columns:
                        if not column in name_list:
                            del station_data[column]

                    
                    station_data = station_data.fillna(0.0001)
                    station_data = station_data.replace('--', 0.0001)       # replace -- symbol to 0
                    station_data_t = station_data_t.fillna(0.0001)
                    station_data_t = station_data_t.replace('--', 0.0001)       # replace -- symbol to 0


                    train_x = train_x.append(station_data, ignore_index=True, sort=False)
                    train_y = train_y.append(station_data_t, ignore_index=True, sort=False)

                    
                print(train_x.shape, train_y.shape, file_count)
                if file_count == during-1:
                    print("Combining data")
                    # curr_idx = None
                    # predict_idx = None
                    # new_train_x = pd.DataFrame(columns=name_list)
                    # new_train_y = pd.DataFrame(columns = self.test_feature_list)

                    full_train_date = []

                    start_date_time = datetime.strptime(curr_year+start_date, '%Y%m%d')
                    curr_date_time_date = start_date_time
                    for d in range(during):
                        curr_date_time_date = start_date_time + timedelta(days=d)
                        day_of_year = curr_date_time_date.date()-datetime.strptime(curr_year, '%Y').date()
                        for h in range(24):
                            curr_date_time_hour = curr_date_time_date + timedelta(hours=h)
                            for m in range(60):
                                curr_date_time_minute = curr_date_time_hour + timedelta(minutes=m)
                                full_train_date.append([curr_date_time_minute.year,curr_date_time_minute.day, day_of_year.days+1, curr_date_time_minute.hour, curr_date_time_minute.minute])
                    
                    full_train_date_df = pd.DataFrame(full_train_date, columns=['Year','DayOfMon','DayOfYear','Hour','Minute'])

                    load_train_date = train_x.loc[:,'Year':'Minute']

                    load_train_date_y = pd.concat([load_train_date, train_y], axis=1, sort=False)   # after average

                    full_train_x = pd.merge(full_train_date_df, train_x,  how='left', on=['Year','DayOfMon','DayOfYear','Hour','Minute'])

                    full_train_y = pd.merge(full_train_date_df, load_train_date_y,  how='left', on=['Year','DayOfMon','DayOfYear','Hour','Minute'])


                    full_train_x.fillna(method='ffill', inplace=True)
                    full_train_y.fillna(method='ffill', inplace=True)

                    print(full_train_x)


                    train_x_date = full_train_x.loc[::5,'Year':'Minute']
                    train_x_average = full_train_x.loc[:,self.feature_list_full_value].groupby(np.arange(len(full_train_x))//seq).mean()

                    # train_x_average.reset_index(drop=True, inplace=True)

                    # train_x_average = pd.concat([train_x_date, train_x_value], axis=1, sort=False)   # after average

                    train_y_average = full_train_y.loc[:,self.test_feature_list].groupby(np.arange(len(full_train_y))//seq).mean()
                    # train_y_average.reset_index(drop=True, inplace=True)

                    count_row = train_x_average.shape[0]
                    train_x_average = train_x_average.head(count_row-6*(60//seq))
                    train_x_average.reset_index(drop=True, inplace=True)
                    train_y_average = train_y_average.head(count_row-6*(60//seq))
                    train_y_average.reset_index(drop=True, inplace=True)

                    train_x_all = pd.concat([train_x_all, train_x_average], ignore_index=True, sort=False)
                    train_y_all = pd.concat([train_y_all, train_y_average], ignore_index=True, sort=False)


                    if not last_year:
                        break
                    else:
                        train_x_all = train_x_all.fillna(0.0001)
                        train_x_all = train_x_all.replace('--', 0.0001)
                        train_y_all = train_y_all.fillna(0.0001)
                        train_y_all = train_y_all.replace('--', 0.0001)


                        print(train_x_all.shape,train_y_all.shape)
                        return train_x_all, train_y_all
                file_count += 1
                

        return None, None
      

    def pre_seq(self, x, y, encoder_seq = 12,  decoder_seq = 6):

        input_texts = x
        target_texts = y
        input_len = len(x)

        max_encoder_seq_len = encoder_seq
        max_decoder_seq_len = decoder_seq+1
        # num_encoder_len = len(x[0])
        # num_decoder_len = len(y[0])

        num_encoder_shape = x.shape
        num_encoder_len = num_encoder_shape[1]  #  TODO: one more control


        num_decoder_len = 1+2  # 2 control bit, -2: start, -1: end

        encoder_input_data = np.zeros((input_len, max_encoder_seq_len, num_encoder_len), dtype='float32')
        decoder_input_data = np.zeros((input_len, max_decoder_seq_len, num_decoder_len), dtype='float32')
        decoder_target_data = np.zeros((input_len, max_decoder_seq_len, num_decoder_len), dtype='float32')


        eff_input_len = 0
        for i, (input_text, target_text) in enumerate(zip(input_texts, target_texts)):

            for t in range(0, max_encoder_seq_len):
                if i+t+encoder_seq < input_len:
                    if t == 0 and i+t+encoder_seq*2 < input_len:
                        eff_input_len += 1

                    encoder_input_data[i, t, ] = input_texts[i+t]


            for t in range(0, max_decoder_seq_len):
                if i+t+encoder_seq < input_len:
                    if t == 0:
                        decoder_input_data[i, t, -2] = 1
                        decoder_target_data[i, t, 0] = target_texts[i+t+encoder_seq]

                    else:
                        decoder_input_data[i, t, 0] = target_texts[i+t+encoder_seq-1]
                        decoder_target_data[i, t, 0] = target_texts[i+t+encoder_seq]


            if i+t+encoder_seq+1 < input_len:
                decoder_target_data[i, t, -1] = 1
        

        encoder_input_data = encoder_input_data[:eff_input_len]
        decoder_input_data = decoder_input_data[:eff_input_len]
        decoder_target_data = decoder_target_data[:eff_input_len]
        
        return encoder_input_data, decoder_input_data, decoder_target_data

 

    def draw_lstm_without_e(self, tru, pre, scaler, lstm_path = None):
        '''
        draw predciton and ground dataset with standard deviation
        '''

        tru = scaler.inverse_transform(tru)
        pre = scaler.inverse_transform(pre)

        idx = 0
        test_len = 2400
        tru_10 = []
        pre_10 = []
        for i in range(min(len(tru),test_len)):
            tru_10.append(tru[i][idx])
            pre_10.append(pre[i][idx])

        # pickle.dump(tru, open("true.p", "wb"))
        pickle.dump(pre, open(lstm_path+'.pre', "wb"))

        fig, ax1 = plt.subplots()

        x = [i*5 for i in range(0, len(tru_10))]

        plt.plot(x, tru_10, label='true')
        plt.plot(x, pre_10, label='prediction')

        plt.ylabel("P_Temp")
        plt.xlabel("minutes")        

        plt.legend()
        plt.show() 

